

#include <zephyr.h>
#include <device.h>
#include <gpio.h>
#include <misc/util.h>
#include <misc/printk.h>

#include <button.h>

#define BUTTON1_PORT  SW0_GPIO_CONTROLLER
#define BUTTON1       SW0_GPIO_PIN

#define BUTTON2_PORT  SW1_GPIO_CONTROLLER
#define BUTTON2       SW1_GPIO_PIN

#define EDGE    (GPIO_INT_EDGE | GPIO_INT_ACTIVE_LOW)

/* change this to enable pull-up/pull-down */
//#ifndef SW0_GPIO_FLAGS
//#ifdef SW0_GPIO_PIN_PUD
//#define SW0_GPIO_FLAGS SW0_GPIO_PIN_PUD
//#else
//#define SW0_GPIO_FLAGS 0
//#endif
//#endif
#define PULL_UP SW0_GPIO_FLAGS

// GPIO setup for button pins.
struct device* bdev1;
struct device* bdev2;

static struct gpio_callback cb1;
static struct gpio_callback cb2;

// State and timing variables for detecting button presses and debouncing.
int state1;
int state2;

s64_t t_change1;
s64_t t_change2;

#define BUTTON_IDLE 1
#define BUTTON_DEBOUNCE_PRESS 2
#define BUTTON_PRESSED 3
#define BUTTON_DEBOUNCE_RELEASE 4

#define debounce_time 10


/*
    Returns pointer to device corresponding to the cardinal number of the button.
*/
struct device* get_button_device(int button_num){
    switch(button_num){
        case 1:
          return bdev1;
        case 2:
          return bdev2;
        default:
          return bdev1;
    }
}

/*
    Returns #define'd PIN corresponding to the cardinal number of the button.
*/
int get_button_pin(int button_num){
    switch(button_num){
        case 1:
          return BUTTON1;
        case 2:
          return BUTTON2;
        default:
          return BUTTON1;
      }
}

/*
    Returns callback struct corresponding to the cardinal number of the button.
*/
struct gpio_callback get_button_cb(int button_num){
    switch(button_num){
        case 1:
          return cb1;
        case 2:
          return cb2;
        default:
          return cb1;
    }
}

/*
    Initializes GPIO for buttons. 
*/
void button_init(int button_num){
    struct device* dev = get_button_device(button_num);
    int pin = get_button_pin(button_num);
    gpio_pin_configure(dev, pin, GPIO_DIR_IN | GPIO_INT | PULL_UP | EDGE);
}

/*
    Initialized both buttons and their respective data structs, GPIO pins, and their state variables. Must be called before other functions can be used.
*/
void button_setup(void){
    bdev1 = device_get_binding(BUTTON1_PORT);
    bdev2 = device_get_binding(BUTTON2_PORT);
    state1 = BUTTON_IDLE;
    state2 = BUTTON_IDLE;
    button_init(1);
    button_init(2);
}

/*
    Returns digital value for the given button. Normal high (1 if button unpressed, 0 if pressed).
*/
u32_t read_button(int button_num){
    struct device* dev = get_button_device(button_num);
    int pin = get_button_pin(button_num);
    u32_t out;
    gpio_pin_read(dev, pin, &out);
    return out;
}

/*
    Returns 1 if the given button is pressed, 0 otherwise.
*/
u32_t button_is_pressed(int button_num){
    return !read_button(button_num);
}

/*
    Sets up function callback to be activated when button is pressed?
*/
void set_button_callback(int button_num, gpio_callback_handler_t cb_function){
    struct device* dev = get_button_device(button_num);
    int pin = get_button_pin(button_num);
    struct gpio_callback cb = get_button_cb(button_num);

    gpio_init_callback(&cb, cb_function, BIT(pin));
    gpio_add_callback(dev, &cb);
    gpio_pin_enable_callback(dev, pin);
}



int* get_button_state(int button_num){
    switch(button_num){
        case 1:
          return &state1;
        case 2:
          return &state2;
        default:
          return &state1;
    }
}

s64_t* get_t_change(int button_num){
    switch(button_num){
        case 1:
          return &t_change1;
        case 2:
          return &t_change2;
        default:
          return &t_change1;
    }
}


int detect_press(int button_num){
    int* BUTTON_STATE = get_button_state(button_num);
    u32_t button_val = read_button(button_num);
    s64_t* t_change = get_t_change(button_num);

    switch(*BUTTON_STATE){
        case BUTTON_IDLE:
          if(button_val == 0){
              *t_change = k_uptime_get();
              *BUTTON_STATE = BUTTON_DEBOUNCE_PRESS;
          }
          break;
        case BUTTON_DEBOUNCE_PRESS:
           if(k_uptime_get() - *t_change >= debounce_time){
              *BUTTON_STATE = BUTTON_PRESSED;
           }else if(button_val == 1){
              *BUTTON_STATE = BUTTON_IDLE;
           }
           break;
        case BUTTON_PRESSED:
          if(button_val == 1){
              *t_change = k_uptime_get();
              *BUTTON_STATE = BUTTON_DEBOUNCE_RELEASE;
          }
          break;
        case BUTTON_DEBOUNCE_RELEASE:
          if(k_uptime_get() - *t_change >= debounce_time){
            *BUTTON_STATE = BUTTON_IDLE;
            return 1;
          }else if(button_val == 0){
            *BUTTON_STATE = BUTTON_PRESSED;
          }
          break;
    }
    return 0;
}




// OOP approaches that weren't friendly to export

//class Button{
//    private:
//    struct device* bdev;
//    struct gpio_callback cb;
//    int pin;
//    u8_t BUTTON_STATE;
//    s64_t t_change;
//    s64_t debounce_time = 10;
//
//    public:
//    Button(const char* PORT, int PIN){
//        pin = PIN;
//        bdev = device_get_binding(PORT);
//        gpio_pin_configure(bdev, pin, GPIO_DIR_IN | GPIO_INT | PULL_UP | EDGE);
//    }
//    u32_t read_state(){
//        // 0 if pressed, 1 if unpressed.
//        u32_t value;
//        gpio_pin_read(bdev, pin, &value);
//        return value;
//    }
//    void set_callback(gpio_callback_handler_t cb_function){
//        gpio_init_callback(&cb, cb_function, BIT(pin));
//        gpio_add_callback(bdev, &cb);
//        gpio_pin_enable_callback(bdev, pin);
//    }
//    int press_detected(){
//        u32_t button_val = read_state();
//        switch(BUTTON_STATE){
//            case BUTTON_IDLE:
//              if(button_val == 0){
//                  t_change = k_uptime_get();
//                  BUTTON_STATE = BUTTON_DEBOUNCE_PRESS;
//              }
//              break;
//            case BUTTON_DEBOUNCE_PRESS:
//               if(k_uptime_get() - t_change >= debounce_time){
//                  BUTTON_STATE = BUTTON_PRESSED;
//               }else if(button_val == 1){
//                  BUTTON_STATE = BUTTON_IDLE;
//               }
//               break;
//            case BUTTON_PRESSED:
//              if(button_val == 1){
//                  t_change = k_uptime_get();
//                  BUTTON_STATE = BUTTON_DEBOUNCE_RELEASE;
//              }
//              break;
//            case BUTTON_DEBOUNCE_RELEASE:
//              if(k_uptime_get() - t_change >= debounce_time){
//                BUTTON_STATE = BUTTON_IDLE;
//                return 1;
//              }else if(button_val == 0){
//                BUTTON_STATE = BUTTON_PRESSED;
//              }
//              break;
//        }
//        return 0;
//    }
//
//}

//typedef struct {
//    struct device* bdev;
//    struct gpio_callback cb;
//    int pin;
//    u8_t STATE;
//    s64_t t_val_change;
//  
//
//
//} Button;
