

#include <zephyr.h>
#include <bsd.h>
#include <nrf_socket.h>
#include <net/socket.h>
#include <stdio.h>

#include <misc/reboot.h>
#include <device.h>
#include <string.h>
#include <flash.h>
#include <nvs/nvs.h>

#include <time.h>
#include <test_locations.h>
#include <led.h>

#include <String.h>

#include <gps_.h>

#define AT_XSYSTEMMODE "AT\%XSYSTEMMODE=0,0,1,0"
#define AT_MAGPIO      "AT\%XMAGPIO=1,0,0,1,1,1574,1577"
#define AT_CFUN        "AT+CFUN=1"

const char     update_indicator[] = {'\\', '|', '/', '-'};
const char     at_commands[][31]  = { AT_XSYSTEMMODE, AT_MAGPIO, AT_CFUN };

char           nmea_strings[10][NRF_GNSS_NMEA_MAX_LEN];
static u32_t   nmea_string_cnt;

int gps_socket;
bool           got_first_fix;
u64_t          fix_timestamp;
nrf_gnss_data_frame_t last_fix;
nrf_gnss_data_frame_t gps_data;

bool do_uart_output;
bool log_data;
int cnt = 0;
s64_t t_last_ping;

bool spoof_gps_data;

u16_t loc_id; // The "id" referencing the next place in flash to record location data
u16_t gps_state;
s64_t gps_update_rate;
s64_t t_last_location_recorded; // Time since bootup that the last location was recorded in flash


//-------------------------------------------------------------------------------------------------------- Prints and Error Handling


void bsd_recoverable_error_handler(uint32_t error)
{
	printf("Err: %lu\n", error);
}

 void bsd_irrecoverable_error_handler(uint32_t error)
 {
 	printf("Irrecoverable: %lu\n", error);
 }


void print_satellite_stats(nrf_gnss_data_frame_t* pvt_data)
{
        /*
          Used to print data when fix not found.
        */
	u8_t  tracked          = 0;
	u8_t  in_fix           = 0;
	u8_t  unhealthy        = 0;

	for (int i = 0; i < NRF_GNSS_MAX_SATELLITES; ++i) {

		if ((pvt_data->pvt.sv[i].sv > 0) &&
		    (pvt_data->pvt.sv[i].sv < 33)) {

			tracked++;

			if (pvt_data->pvt.sv[i].flags &
					NRF_GNSS_PVT_FLAG_FIX_VALID_BIT) {
				in_fix++;
			}

			if (pvt_data->pvt.sv[i].flags &
					NRF_GNSS_SV_FLAG_UNHEALTHY) {
				unhealthy++;
			}
		}
	}

	printk("Tracking: %d Using: %d Unhealthy: %d", tracked,
						       in_fix,
						       unhealthy);

	printk("\nSeconds since last fix %lld\n",
			(k_uptime_get() - fix_timestamp) / 1000);
}

void print_pvt_data(nrf_gnss_data_frame_t *pvt_data)
{
	printf("Longitude:  %f\n", pvt_data->pvt.longitude);
	printf("Latitude:   %f\n", pvt_data->pvt.latitude);
	printf("Altitude:   %f\n", pvt_data->pvt.altitude);
	printf("Speed:      %f\n", pvt_data->pvt.speed);
	printf("Heading:    %f\n", pvt_data->pvt.heading);
	printk("Date:       %02u-%02u-%02u\n", pvt_data->pvt.datetime.day,
					       pvt_data->pvt.datetime.month,
					       pvt_data->pvt.datetime.year);
	printk("Time (UTC): %02u:%02u:%02u\n", pvt_data->pvt.datetime.hour,
					       pvt_data->pvt.datetime.minute,
					      pvt_data->pvt.datetime.seconds);
}

void print_store_data(nrf_gnss_data_frame_t *gps_data){

    	printf("Longitude:  %f\n", gps_data->pvt.longitude);
	printf("Latitude:   %f\n", gps_data->pvt.latitude);

}

void print_nmea_data(void)
{
	printk("NMEA strings:\n");

	for (int i = 0; i < nmea_string_cnt; ++i) {
		printk("%s\n", nmea_strings[i]);
	}
}


//---------------------------------------------------------------------------------------------------------- GPS Processing


int enable_gps(void)
{
 	int  at_sock;
 	int  bytes_sent;
 	int  bytes_received;
 	char buf[2];

 	at_sock = socket(AF_LTE, 0, NPROTO_AT);
 	if (at_sock < 0) {
                printk("Bad at_sock");
 		return -1;
 	}

 	for (int i = 0; i < ARRAY_SIZE(at_commands); i++) {
 		bytes_sent = send(at_sock, at_commands[i],
 				  strlen(at_commands[i]), 0);

 		if (bytes_sent < 0) {
                        printk("Negative Bytes");
 			close(at_sock);
 			return -1;
 		}

 		do {
 			bytes_received = recv(at_sock, buf, 2, 0);
 		} while (bytes_received == 0);

 		if (memcmp(buf, "OK", 2) != 0) {
                        printk("Not OK");
 			close(at_sock);
 			return -1;
 		}
 	}

 	close(at_sock);

 	return 0;
 }

int init_app(void)
{
        /*
            Initializes GPS settings and socket for GPS comms.
        */
	u16_t fix_retry     = 0;
	u16_t fix_interval  = 1;
	u16_t nmea_mask     = NRF_CONFIG_NMEA_GSV_MASK |
			      NRF_CONFIG_NMEA_GSA_MASK |
			      NRF_CONFIG_NMEA_GLL_MASK |
			      NRF_CONFIG_NMEA_GGA_MASK |
			      NRF_CONFIG_NMEA_RMC_MASK;
	int   retval;

	if (enable_gps() != 0) {
		printk("Failed to enable GPS\n");
		return -1;
	}

	gps_socket = nrf_socket(NRF_AF_LOCAL, NRF_SOCK_DGRAM, NRF_PROTO_GNSS);

	if (gps_socket >= 0) {
		printk("Socket created\n");
	} else {
		printk("Could not init socket (err: %d)\n", gps_socket);
		return -1;
	}

	retval = nrf_setsockopt(gps_socket,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_FIX_RETRY,
				&fix_retry,
				sizeof(uint16_t));

	if (retval != 0) {
		printk("Failed to set fix retry value\n");
		return -1;
	}

	retval = nrf_setsockopt(gps_socket,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_FIX_INTERVAL,
				&fix_interval,
				sizeof(uint16_t));

	if (retval != 0) {
		printk("Failed to set fix interval value\n");
		return -1;
	}

	retval = nrf_setsockopt(gps_socket,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_NMEA_MASK,
				&nmea_mask,
				sizeof(uint16_t));

	if (retval != 0) {
		printk("Failed to set nmea mask\n");
		return -1;
	}

	retval = nrf_setsockopt(gps_socket,
				NRF_SOL_GNSS,
				NRF_SO_GNSS_START,
				NULL,
				0);

	if (retval != 0) {
		printk("Failed to start GPS\n");
		return -1;
	}

	return 0;
}

int process_gps_data(nrf_gnss_data_frame_t *gps_data)
{
	int retval;

	retval = nrf_recv(gps_socket, gps_data, sizeof(nrf_gnss_data_frame_t), NRF_MSG_DONTWAIT);

	if (retval > 0) {

		switch (gps_data->data_id) {
		case NRF_GNSS_PVT_DATA_ID:

			if ((gps_data->pvt.flags &
				NRF_GNSS_PVT_FLAG_FIX_VALID_BIT)
				== NRF_GNSS_PVT_FLAG_FIX_VALID_BIT) {

				if (!got_first_fix) {
					got_first_fix = true;
				}

				fix_timestamp = k_uptime_get();
				memcpy(&last_fix, gps_data, sizeof(nrf_gnss_data_frame_t));

				nmea_string_cnt = 0;
			}
			break;

		case NRF_GNSS_NMEA_DATA_ID:
			if (nmea_string_cnt < 10) {
				memcpy(nmea_strings[nmea_string_cnt++],
				       gps_data->nmea,
				       retval);
			}
			break;

		default:
			break;
		}
	}

	return retval;
}


void display_gps_fixed(nrf_gnss_data_frame_t* gps_data){
        /*
            Displays data from a GPS recording.
        */
	printk("\033[1;1H");
	printk("\033[2J");

	print_satellite_stats(gps_data);

	printk("---------------------------------\n");
	print_pvt_data(&last_fix);
	printk("\n");
	print_nmea_data();
	printk("---------------------------------");
}


int get_gps_fix(nrf_gnss_data_frame_t* gps_data){
    /*
        Records data from gps and prints it out every 500ms until a fix is detected.
    */

    if(spoof_gps_data) return 0;

    do {
    
    } while(process_gps_data(gps_data) > 0); // Read data from GPS
     
    if(!got_first_fix){
        cnt = (cnt+1)%4;
        printk("\033[1;1H");
        printk("\033[2J");
        print_satellite_stats(gps_data);
        printk("\nScanning: [%c] ", update_indicator[cnt]);
    }
    else if(k_uptime_get()-fix_timestamp >= 1){
                    display_gps_fixed(gps_data);
    }

    return 0;
}

unsigned long int location_date_to_epoch(nrf_gnss_data_frame_t *gps_data){
    /*
        Returns the time recorded in a gps_data struct as an epoch time number (GMT)
        for easier storage and handling.
    */

    nrf_gnss_datetime_t dt = gps_data->pvt.datetime;
    // Copy date data from the GPS struct into the tm struct so it can be converted.
    struct tm t;
    t.tm_year = dt.year;
    t.tm_mon = dt.month;
    t.tm_mday = dt.day;
    t.tm_hour = dt.hour;
    t.tm_min = dt.minute;
    t.tm_sec = dt.seconds;
    t.tm_isdst = -1;

    // Convert to unsigned long and add offset
    return (unsigned long int)(mktime(&t)) + 168719744;

}

int data_ix;
void update_gps_data(nrf_gnss_data_frame_t *gps_data){
        /*
            This f'n populates gps_data with new information coming from the gps.
            For debugging, it simply runs through some spoofed GPS data.
        */
        if(do_uart_output) printk("\nUpdating GPS Data\n");

        if(spoof_gps_data){
            if(data_ix == 1){
                *gps_data = bradford;
            }else if(data_ix == 2){
                *gps_data = random;
                data_ix = 0;
            }else{
                *gps_data = office;
            }  

            data_ix++;

            if(do_uart_output) print_pvt_data(gps_data);
            last_fix = *gps_data;
            return;
        }


        do {
        
        } while(process_gps_data(gps_data) > 0); // Read data from GPS

        if(do_uart_output) print_pvt_data(&last_fix);
}

void save_gps_data(nrf_gnss_data_frame_t *gps_data, struct nvs_fs* fs){
        /*
            This f'n saves the data held in gps_data to the flash storage and
            increments the "id" to a blank set of spaces to record the next set of data.
        */

        printk("\nWriting Coords\n");
	nrf_gnss_pvt_data_frame_t pvt = gps_data->pvt;

//        printk("LAT ");
	nvs_write(fs, loc_id, &pvt.latitude, (size_t)sizeof(double));
//        printk("LON ");
	nvs_write(fs, loc_id+1, &pvt.longitude, (size_t)sizeof(double));
//        printk("ACC ");
	nvs_write(fs, loc_id+2, &pvt.accuracy, (size_t)sizeof(float));
//        printk("DATE ");
        unsigned long loc_time = location_date_to_epoch(gps_data);
	nvs_write(fs, loc_id+3, &loc_time, (size_t)sizeof(unsigned long));
	
        loc_id+=4;
        
//        printk("\nDone Writing\n");
        printk("Locations: %d; Location ID: %d \n", (loc_id-1)/4, loc_id);
}


//-------------------------------------------------------------------------------------------------------- Outside-accessable functions


s64_t GPS_Update_Rate(s64_t new_rate){
    /*
        Updates the rate at which the device should save GPS data to flash.
    */
    if(new_rate >= 0) gps_update_rate = new_rate;
    return gps_update_rate;
}


void GPS_Logging_On(){
    log_data = true;
}

void GPS_Logging_Off(){
    log_data = false;
}

bool GPS_Is_Logging(){
    return log_data;
}

void UART_Output_On(){
    do_uart_output = true;
}

void UART_Output_Off(){
    do_uart_output = false;
}


void spoof(bool set){
    spoof_gps_data = set;
}


void GPS_Init(){
    /*
        Sets default variable values and sets the socket options for the GPS.
    */
    
    UART_Output_On();
    GPS_Logging_On();
    GPS_Update_Rate(900); 
    loc_id = 1;
    if(spoof_gps_data){
        gps_data = office;
        gps_state = GPS_IDLE;
        UART_Output_Off();
    }else{
        gps_state = GET_FIX;
    }
    t_last_location_recorded = k_uptime_get();
    
}

int GPS_Enable(){
    if(!spoof_gps_data){
        int rv = init_app();
        if(rv) printk("init_app failure: rv = %d", rv);
    }

    return get_gps_fix(&gps_data);
}

int GPS_Disable(){
    return nrf_close(gps_socket);
}

double Get_Lat(struct nvs_fs* fs){
    /*
        Return last latitude saved in memory. If none saved, return 0.
    */
    if(loc_id<=1) return 0;

    double last_lat;
    nvs_read(fs, loc_id-4, &last_lat, sizeof(double));
    return last_lat;
}

double Get_Lon(struct nvs_fs* fs){
    /*
        Return last longitude saved in memory. If none saved, return 0.
    */
    if(loc_id<=1) return 0;

    double last_lon;
    nvs_read(fs, loc_id-3, &last_lon, sizeof(double));
    return last_lon;
}

unsigned long Get_Time(struct nvs_fs* fs){
    /*
        Return last location time saved in memory. If none saved, return 0.
    */
    if(loc_id<=1) return 0;

    unsigned long last_time;
    nvs_read(fs, loc_id-1, &last_time, sizeof(unsigned long));
    return last_time;
}

void Dump_GPS_Coordinates(struct nvs_fs* fs){
        /*
            Print all recorded GPS data in a csv string over UART.
        */
        printk("\nSending Coords:\n");
        double last_lat;
        double last_lon;
        float last_acc;
        unsigned long last_time;
        for(int last_loc_id = 1; last_loc_id <= loc_id-4; last_loc_id += 4){
            nvs_read(fs, last_loc_id, &last_lat, sizeof(double));
            printf("%f,", last_lat);
            nvs_read(fs, last_loc_id+1, &last_lon, sizeof(double));
            printf("%f,", last_lon);
            nvs_read(fs, last_loc_id+2, &last_acc, sizeof(float));
            printf("%f,", last_acc);
            nvs_read(fs, last_loc_id+3, &last_time, sizeof(unsigned long));
            printf("%lu\n", last_time);
        }
        printk("Done Sending\n");

}


//struct coordinate_data {
//    double latitude;
//    double longitude;
//    float accuracy;
//    unsigned long timestamp;
//};




struct coordinate_data make_coord_struct(struct nvs_fs* fs, int id){

        struct coordinate_data buffer_coord;

        double last_lat;
        double last_lon;
        float last_acc;
        unsigned long last_time;

        nvs_read(fs, id, &last_lat, sizeof(double));
        nvs_read(fs, id+1, &last_lon, sizeof(double));
        nvs_read(fs, id+2, &last_acc, sizeof(float));
        nvs_read(fs, id+3, &last_time, sizeof(unsigned long));

        buffer_coord.latitude = last_lat;
        buffer_coord.longitude = last_lon;
        buffer_coord.accuracy = last_acc;
        buffer_coord.timestamp = last_time;
        
        return buffer_coord;
}

void make_coord_json_string(char* out, struct nvs_fs* fs, int id){
          
      struct coordinate_data coord = make_coord_struct(fs, id);  
      sprintf(out, "{\"latitude\": %f,\"longitude\": %f,\"accuracy\": %f,\"timestamp\": %lu,},", coord.latitude, coord.longitude, coord.accuracy, coord.timestamp);

}




void make_coordinate_struct_array(struct nvs_fs* fs, struct coordinate_data* out){

        double last_lat;
        double last_lon;
        float last_acc;
        unsigned long last_time;

        struct coordinate_data buffer_coord;

        for(int last_loc_id = 1; last_loc_id <= loc_id-4; last_loc_id += 4){
            
            nvs_read(fs, last_loc_id, &last_lat, sizeof(double));
            nvs_read(fs, last_loc_id+1, &last_lon, sizeof(double));
            nvs_read(fs, last_loc_id+2, &last_acc, sizeof(float));
            nvs_read(fs, last_loc_id+3, &last_time, sizeof(unsigned long));

            buffer_coord.latitude = last_lat;
            buffer_coord.longitude = last_lon;
            buffer_coord.accuracy = last_acc;
            buffer_coord.timestamp = last_time;

            out[(last_loc_id-1)/4] = buffer_coord;
        }
}




char* make_coord_array_json_string(char* out, struct coordinate_data* array, int size){
    strcat(out, "{\"coordinates\":[");

    struct coordinate_data* coord;
    char buffer[100];

    for(int ix=0; ix < size; ix++){
        coord = &array[ix];

        memset(buffer, 0, sizeof(buffer));
        sprintf(buffer, "{\"latitude\": %f,\"longitude\": %f,\"accuracy\": %f,\"timestamp\": %lu},", coord->latitude, coord->longitude, coord->accuracy, coord->timestamp);
        strcat(out, buffer);

    }

    strcat(out, "]}");
    return out;
}

int Get_Loc_ID(){
    return loc_id;
}


void GPS_State_Machine(struct nvs_fs* fs){
    switch(gps_state){
		case GPS_IDLE:
                    
                    // Record a coordinate every gps_update_rate period
                    if(k_uptime_get()-t_last_location_recorded >= GPS_Update_Rate(-1)){
                        t_last_location_recorded = k_uptime_get();
                        gps_state = RECORD_COORDINATE;
                    }
                    break;

		case RECORD_COORDINATE:{

                    // Record a location in memory
                    toggle_led(1);
                    update_gps_data(&gps_data);

                    if(log_data){
                        save_gps_data(&last_fix, fs);
                    }

                    toggle_led(1);

                    if(do_uart_output) Dump_GPS_Coordinates(fs);

//                    char strprint[100];
//                    make_coord_json_string(strprint, fs, loc_id-4);
//                    printk("\n%s\n", strprint);


//                    int num_of_locs = (loc_id-1)/4;
//                    struct coordinate_data coords_data[num_of_locs];
//                    make_coordinate_struct_array(fs, coords_data);
//                    char bigprint[100*num_of_locs];
//                    make_coord_array_json_string(bigprint, coords_data, num_of_locs);
//                    printk("%s\n", bigprint);


//                        printf("\n\n tllr: %llu \n\n", t_last_location_recorded);
                    gps_state = GPS_IDLE;
                    break;}

                case GET_FIX:
                    // Ping every FIX_PING_RATE period to try to get a fix
                    if(k_uptime_get()-t_last_ping >= FIX_PING_RATE){
                        t_last_ping = k_uptime_get();
                        toggle_led(2);

                        if(!got_first_fix){
                            get_gps_fix(&gps_data);
                        }
                        if(got_first_fix){
                            led_on(2);
                            gps_state = GPS_IDLE;
                        }

                    }
                    break;
	}
}




